{*******************************************************}
{          Copyright (c) 2000 Vlad Korshunov            }
{                   IPH tools project                   }
{              Common plugin implementation             }
{*******************************************************}
unit Plugin;
interface
uses windows, sysutils, PluginAPI, probes, classes, ProbeParams, utils;

type

  TDLLProbe = class
  private
    HWND: HWND;
    FProbe: TCustomProbe;
    FMsg: Integer;
    procedure ProbeNotify(Probe: TCustomProbe);
  public
    constructor Create;
    destructor Destroy; override;
  end;


function GetProbeInfo(ProbeName: PChar; Description: PChar;
  var BufSize: Integer): Boolean; stdcall;

procedure EnumProbeNames(cbProbeName: TProbeNameCallback; Data: Pointer); stdcall;
procedure EnumProbeParams(ProbeName: PChar; cbProbeParams: TProbeParamCallback;
    Data: Pointer); stdcall;
function StartProbe(ProbeName: PChar; cbParamValues: TcbGetValue; Msg: Integer;
  HWND: HWND; Data: Pointer): THandle; stdcall;
procedure StopProbe(ProbeHandle: Integer); stdcall;
procedure GetCurrentValues(ProbeHandle: Integer;
  cbParamValues: TcbSetValue; Data: Pointer); stdcall;

implementation
const
  FAvailableProbes: TStrings = nil;
  FProbes: TList = nil;

procedure InitAvailableProbes;
begin
  if FAvailableProbes = nil then
  begin
    FAvailableProbes:=TStringList.Create;
    GetAvailableProbes(FAvailableProbes);
  end;
end;

function GetProbeClassByName(ProbeName: PChar): TProbeClass;
var
  Index: integer;
begin
  InitAvailableProbes;
  if ProbeName = nil then
  begin
    if FAvailableProbes.Count=0 then
      Raise Exception.Create('No available probes');
    Index:=0;
  end else
  begin
    Index:=FAvailableProbes.IndexOf(ProbeName);
    if Index=-1 then
      Raise Exception.CreateFmt('Invalid probe name %s', [ProbeName]);
  end;
  result:=TProbeClass(FAvailableProbes.Objects[Index]);
end;

function GetProbeInfo(ProbeName: PChar; Description: PChar;
  var BufSize: Integer): Boolean; stdcall;
var
  ProbeClass: TProbeClass;
begin
  ProbeClass:=GetProbeClassByName(ProbeName);
  result:=CopyStringToBuf(ProbeClass.GetDescription, Description, BufSize);
end;

procedure EnumProbeNames(cbProbeName: TProbeNameCallback; Data: Pointer); stdcall;
var
  i: integer;
begin
  InitAvailableProbes;
  for i:=0 to FAvailableProbes.Count-1 do
    if not cbProbeName(PChar(FAvailableProbes[i]), Data) then Break;
end;

procedure EnumProbeParams(ProbeName: PChar; cbProbeParams: TProbeParamCallback;
    Data: Pointer); stdcall;
var
  i: integer;
begin
  With GetProbeClassByName(ProbeName).Create do
  try
    for i:=0 to Params.Count-1 do
      With Params[i] do
        cbProbeParams(PChar(Name), Integer(DataType), ParamType = ptInput,
          poRequired in ParamOptions, PChar(AsString), Data);
  finally
    free;
  end;
end;

function StartProbe(ProbeName: PChar; cbParamValues: TcbGetValue; Msg: Integer;
  HWND: HWND; Data: Pointer): THandle; stdcall;
var
  DLLProbe: TDllProbe;
  i: integer;
  Value: String;
  ReqSize: Integer;
begin
  DLLProbe:=TDLLProbe.Create;
  DLLProbe.FProbe:=GetProbeClassByName(ProbeName).Create;
  DLLProbe.FMsg:=Msg;
  DLLProbe.HWND:=HWND;
  DLLProbe.FProbe.NotifyByMessage:=False;
  DLLProbe.FProbe.Notify:=DLLProbe.ProbeNotify;
  for i:=0 to DLLProbe.FProbe.Params.Count-1 do
  begin
    Value:='';
    reqSize:=0;
    if not cbParamValues(PChar(DLLProbe.FProbe.Params[i].Name), Data, @Value[1], reqSize) then
    begin
      SetLength(Value, ReqSize);
      cbParamValues(PChar(DLLProbe.FProbe.Params[i].Name), Data, @Value[1], reqSize);
    end;
    DLLProbe.FProbe.Params[i].AsString:=Value;
  end;
  if FProbes = nil then FProbes:=TList.Create;
  FProbes.Add(DLLProbe);
  DLLProbe.FProbe.StartProbe;
  result:=Integer(DLLProbe);
end;

function DLLProbeByHandle(ProbeHandle: Integer): TDLLProbe;
begin
  if (ProbeHandle<>0) and (TObject(Pointer(ProbeHandle)) is TDLLProbe) then
    result:=TDLLProbe(Pointer(ProbeHandle)) else
      Raise Exception.Create('Invalid probe handle');
end;

procedure StopProbe(ProbeHandle: Integer); stdcall; export;
var
  DLLProbe: TDLLProbe;
begin
  DLLProbe:=DLLProbeByHandle(ProbeHandle);
  DLLProbe.FProbe.StopProbe;
  DLLProbe.Free;
end;

procedure GetCurrentValues(ProbeHandle: Integer; cbParamValues: TcbSetValue;
  Data: Pointer); stdcall; export;
var
  i: integer;
  DLLProbe: TDLLProbe;
  P: PChar;
begin
  DLLProbe:=DLLProbeByHandle(ProbeHandle);
  for i:=0 to DLLProbe.FProbe.Params.Count-1 do
  begin
    p:=PChar(DLLProbe.FProbe.Params[i].AsString);
    cbParamValues(PChar(DLLProbe.FProbe.Params[i].Name), P, Data)
  end;
end;

{ TDLLProbe }

constructor TDLLProbe.Create;
begin
  inherited Create;
end;

destructor TDLLProbe.Destroy;
begin
  FProbes.Remove(Self);
  FProbe.Free;
  inherited;
end;

procedure TDLLProbe.ProbeNotify(Probe: TCustomProbe);
begin
  SendMessage(HWND, FMsg, Integer(Self), 0);
end;

procedure finit;
var
  i: integer;
begin
  if FProbes<>nil then
  begin
    for i:=FProbes.Count-1 downto 0 do
      TDLLProbe(FProbes[i]).Free;
    FProbes.Free;
  end;
  FAvailableProbes.Free;
end;

initialization
finalization
  finit;
end.
