{*******************************************************}
{          Copyright (c) 2000 Vlad Korshunov            }
{                   IPH tools project                   }
{                 Sockets related probes                }
{*******************************************************}
unit SockProbes;
interface
uses windows, messages, sysutils, winsock, probes, probeParams;

Type
  TIPRangeIterator = class(TRangeIterator)
  protected
    function OrdinalForm(ComplexValue: Integer): Integer; override;
    function ComplexForm(OrdinalValue: Integer): Integer; override;
    procedure GetParamNames(var LoValueName, HiValueName, CurValueName: String;
      var ValueType: TFieldType); override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
  end;

  TPortRangeIterator = class(TRangeIterator)
  protected
    procedure GetParamNames(var LoValueName, HiValueName, CurValueName: String;
      var ValueType: TFieldType); override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
  end;

  THostByName = class(TTimeredProbe)
  private
    FDnsLookupBuffer  : array [0..MAXGETHOSTSTRUCT] of char;
    FDnsLookupHandle  : THandle;
  protected
    procedure AddSupportedParams(var Params: TParams); override;
    procedure InternalStartProbe; override;
    procedure InternalStopProbe; override;
    procedure InternalWndProc(var Msg: TMessage); override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
  end;

  THostByAddr = class(TTimeredProbe)
  private
    FDnsLookupBuffer  : array [0..MAXGETHOSTSTRUCT] of char;
    FDnsLookupHandle  : THandle;
  protected
    procedure AddSupportedParams(var Params: TParams); override;
    procedure InternalStartProbe; override;
    procedure InternalStopProbe; override;
    procedure InternalWndProc(var Msg: TMessage); override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
  end;

  THostAddress = class(TActionProbe)
  private
    FHostByName: THostByName;
    procedure HostResolved(Probe: TCustomProbe);
  protected
    procedure AddSupportedParams(var Params: TParams); override;
    function InternalAction: Boolean; override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
    destructor Destroy; override;
  end;

type
  TAsyncStyle = (asRead, asWrite, asOOB, asAccept, asConnect, asClose);
  TAsyncStyles = set of TAsyncStyle;

  TCustomSocketProbe = class(TTimeredProbe)
  private
    FSocket: TSocket;
  protected
    procedure InternalWndProc(var Msg : TMessage); override;
    procedure CloseSocket; virtual;

    function OpenSocket: TSocket; virtual; abstract;
    procedure DoReceive; virtual; abstract;
    procedure DoSend; virtual; abstract;
    procedure DoConnect; virtual; abstract;
    procedure DoClose; virtual; abstract;
    procedure InternalStartSocket; virtual; abstract;

    procedure AddSupportedParams(var Params: TParams); override;
    procedure InternalStartProbe; override;
    procedure InternalStopProbe; override;
    procedure SetAsyncStyles(Socket: TSocket; AsyncStyles: TAsyncStyles);
  public
    constructor Create; override;
    destructor destroy; override;
  end;

  TPingProbe = class(TCustomSocketProbe)
  protected
    function OpenSocket: TSocket; override;
    procedure InternalStartSocket; override;
    procedure DoReceive; override;
    procedure AddSupportedParams(var Params: TParams); override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
  end;

  TTCPPortProbe = class(TCustomSocketProbe)
  protected
    function OpenSocket: TSocket; override;
    procedure InternalStartSocket; override;
    procedure DoReceive; override;
    procedure DoConnect; override;
    procedure AddSupportedParams(var Params: TParams); override;
  public
    class function GetName: string; override;
    class function GetDescription: String; override;
  end;

const
  WM_SOCKET_EVENT       = WM_ACTION+1;
  WM_ASYNCGETHOSTBYNAME = WM_ACTION+2;

implementation
uses iphconsts, socks, sockerr;

{$I ICMP.inc}

{ TIPRangeIterator }

function TIPRangeIterator.ComplexForm(OrdinalValue: Integer): Integer;
begin
  result:=ntohl(OrdinalValue);
end;

class function TIPRangeIterator.GetDescription: String;
begin
  result:=sIPRangeIteratorDesc;
end;

class function TIPRangeIterator.GetName: string;
begin
  result:='IP range iterator';
end;

procedure TIPRangeIterator.GetParamNames(var LoValueName, HiValueName, CurValueName: String;
  var ValueType: TFieldType);
begin
  ValueType:=ftIPAddress;
  LoValueName:='From IP'; HiValueName:='To IP'; CurValueName:='Address';
end;

function TIPRangeIterator.OrdinalForm(ComplexValue: Integer): Integer;
begin
  result:=htonl(ComplexValue);
end;

{ TPortRangeIterator }

class function TPortRangeIterator.GetDescription: String;
begin
  result:=sPortRangeIteratorDesc;
end;

class function TPortRangeIterator.GetName: string;
begin
  result:='Port range iterator';
end;

procedure TPortRangeIterator.GetParamNames(var LoValueName, HiValueName,
  CurValueName: String; var ValueType: TFieldType);
begin
  ValueType:=ftInteger;
  LoValueName:='From port'; HiValueName:='To port'; CurValueName:='Port';
end;

{ TCustomSocketProbe }

procedure TCustomSocketProbe.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftIPAddress, 'Address', ptInput, [poRequired]);
end;

procedure TCustomSocketProbe.CloseSocket;
begin
  winsock.CloseSocket(FSocket);
  FSocket:=INVALID_SOCKET;
end;

constructor TCustomSocketProbe.Create;
begin
  inherited Create;
  FSocket:=INVALID_SOCKET;
end;

destructor TCustomSocketProbe.destroy;
begin
  if FSocket<>INVALID_SOCKET then Winsock.CloseSocket(FSocket);
  inherited;
end;

procedure TCustomSocketProbe.InternalStartProbe;
begin
  inherited;
  InitWinsock;
  FSocket:=OpenSocket;
  InternalStartSocket;
end;

procedure TCustomSocketProbe.InternalStopProbe;
begin
  StopTimer;
  CloseSocket;
  inherited;
end;

procedure TCustomSocketProbe.InternalWndProc(var Msg: TMessage);

  function GetApiName(Style: Integer): String;
  begin
    case Style of
      fd_read: result:='fd_read';
      fd_write: result:='fd_write';
      fd_OOB: result:='fd_OOB';
      fd_accept: result:='fd_accept';
      fd_connect: result:='fd_connect';
      fd_close: result:='fd_close';
      else result:='Unknown';
    end;
  end;

begin
  Case msg.msg of
    WM_SOCKET_EVENT:
    begin
      CheckSocketLastErr(msg.lparamhi, GetApiName(msg.lparamlo));
      case msg.lparamlo of
        fd_read: DoReceive;
        fd_write: DoSend;
        fd_connect: DoConnect;
        fd_close: DoClose;
      end;
    end;
    else inherited;
  end;
end;

procedure TCustomSocketProbe.SetAsyncStyles(Socket: TSocket; AsyncStyles: TAsyncStyles);
begin
  CheckSocketResult(winsock.WSAAsyncSelect(Socket, hwnd, WM_SOCKET_EVENT,
    Longint(Byte(AsyncStyles))), 'WSAAsyncSelect');
end;

{ TPingProbe }

procedure TPingProbe.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftInteger, 'BlockSize', ptInput, []).Value:=64;
//  Params.CreateParam(ftInteger, 'TimeToLive', ptInput, []).Value:=255;

  Params.CreateParam(ftInteger, 'PingTime', ptOutput, []);
  Params.CreateParam(ftIPAddress, 'ReplyIP', ptOutput, []);
end;

class function TPingProbe.GetName: string;
begin
  result:='Ping host';
end;

function TPingProbe.OpenSocket: TSocket;
begin
  result:=GetSocket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
  SetAsyncStyles(result, [asread])
end;

const
  fpacket_no: word = 0;

function now_ms:longint;
var
  systime : TSystemTime;
begin
  GetLocalTime(systime);
  result:=systime.wmilliseconds+1000*systime.wsecond+60000*systime.wminute;
end;

function ICMP_checksum(var buf; length:integer):word;
var
  p: pointer;
  sum: longint;
  i: integer;
begin
  p:=@buf;
  sum:=0;
  for i:=1 to length div 2 do begin
    sum:=sum+word(p^);
    p:=pointer(longint(p)+2);
    end;
  if length mod 1<>0 then
    sum:=sum+byte(p^);

  sum:=(sum shr 16) + (sum and $ffff);
  sum:=sum+(sum shr 16);
  result:=word(NOT sum);
end;

procedure TPingProbe.DoReceive;
var
  replybuffer: pointer;
  ReplySize: dword;
  from: TSockAddrIn;
begin
  replybuffer:=nil;
  ReplySize:=sizeof(t_icmp_echo_request)+sizeof(t_ip_header)+sizeof(longint)+
    Params.ParamByName('BlockSize').AsInteger;
  try
    GetMem(replybuffer,ReplySize);
    from.sin_family:=AF_INET;
    from.sin_port:=0;
    from.sin_addr.S_addr:=Params.ParamByName('Address').AsInteger;
    SocketRead(FSocket, replybuffer^, ReplySize, @from);
    if (t_ip_header(replybuffer^).ip_hl_v=$45) and
      (TICMPreply_buf(replybuffer^).icmp_header.icmp_id=word(self)) then
      begin
        Params.ParamValues['PingTime']:=now_ms-TICMPreply_buf(replybuffer^).time;
        Params.ParamValues['ReplyIP']:=TICMPreply_buf(replybuffer^).ip_header.ip_src;
        DoOnProbe(True);
      end;
  finally
    if replybuffer<>NIL then FreeMem(replybuffer,ReplySize);
  end;
end;

procedure TPingProbe.InternalStartSocket;
var
  requestdata: pointer;
  whereto: TSockAddr;
  rq_size: integer;
begin
  requestdata:=nil;
  rq_size:=sizeof(t_icmp_echo_request)+Params.ParamByName('BlockSize').AsInteger+sizeof(longint);
  try
    GetMem(requestdata,rq_size);
    ticmp_sendblock(requestdata^).icmp_header.icmp_type := ICMP_ECHO;
    ticmp_sendblock(requestdata^).icmp_header.icmp_code := 0;
    ticmp_sendblock(requestdata^).icmp_header.icmp_cksum := 0;
    ticmp_sendblock(requestdata^).icmp_header.icmp_seq := fpacket_no;
    inc(fpacket_no);
    ticmp_sendblock(requestdata^).icmp_header.icmp_id := word(self);
    fillchar(ticmp_sendblock(requestdata^).data,Params.ParamByName('BlockSize').AsInteger,#$a7);
    ticmp_sendblock(requestdata^).time:=now_ms;
    ticmp_sendblock(requestdata^).icmp_header.icmp_cksum := ICMP_checksum(requestdata^,rq_size);
    whereto.sin_family:=AF_INET;
    whereto.sin_port:=0;
    whereto.sin_addr.S_addr:=Params.ParamByName('Address').AsInteger;
    SocketWrite(FSocket, requestdata^, rq_size, @whereto);
  finally
    if requestdata<>NIL then
      FreeMem(requestdata,sizeof(t_icmp_echo_request)+Params.ParamByName('BlockSize').AsInteger);
  end;
end;

class function TPingProbe.GetDescription: String;
begin
  result:=sPingDesc;
end;

{ THostByName }
procedure THostByName.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftString, 'HostName', ptInput, [poRequired]);
  Params.CreateParam(ftIPAddress, 'Address', ptOutput, []);
end;

class function THostByName.GetDescription: String;
begin
  result:=sHostByNameDesc;
end;

class function THostByName.GetName: string;
begin
  Result:='Get host by name';
end;

procedure THostByName.InternalStartProbe;
begin
  InitWinsock;
  FDnsLookupHandle := WSAAsyncGetHostByName(HWND,
    WM_ASYNCGETHOSTBYNAME, PChar(Params.ParamByname('HostName').AsString),
     @FDnsLookupBuffer, SizeOf(FDnsLookupBuffer));
  CheckSocketResult(Ord(FDnsLookupHandle = 0), 'WSAASyncGetHostByName');
end;

procedure THostByName.InternalStopProbe;
begin
  CheckSocketResult(WSACancelASyncRequest(FDnsLookupHandle), 'WSACancelASyncRequest');
  inherited;
end;

procedure THostByName.InternalWndProc(var Msg: TMessage);
var
    Phe     : Phostent;
begin
  if (Msg.Msg <> WM_ASYNCGETHOSTBYNAME) or (msg.wParam <> LongInt(FDnsLookupHandle)) then
    inherited;
  FDnsLookupHandle := 0;
  if Msg.LParamHi<>0 then
    raise ESocketError.CreateFmt(sWindowsSocketError,
      [GetErrorDescription(Msg.LParamHi), Msg.LParamHi, 'WSAASyncGetHostByName']);
  Phe := PHostent(@FDnsLookupBuffer);
  if (Phe<>nil) and (Phe^.h_addr_list<>nil) then
  begin
    Params.ParamByName('Address').AsInteger:=Integer(PInAddr(Phe^.h_addr_list^)^);
    DoOnProbe(True);
  end else
    DoOnError(0, sNoDNSEntryFound);
end;

{ THostByAddr }

procedure THostByAddr.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftIPAddress, 'Address', ptInput, [poRequired]);
  Params.CreateParam(ftString, 'HostName', ptOutput, []);
end;

class function THostByAddr.GetDescription: String;
begin
  result:=sHostByAddrDesc;
end;

class function THostByAddr.GetName: string;
begin
  result:= 'Get host name';
end;

procedure THostByAddr.InternalStartProbe;
var
  Addr: TInAddr;
begin
  Addr:=in_Addr(Params.ParamByName('Address').AsInteger);
  InitWinsock;
  FDnsLookupHandle := WSAAsyncGetHostByAddr(HWND,
    WM_ASYNCGETHOSTBYNAME, PChar(@Addr), Sizeof(Addr), af_Inet,
     @FDnsLookupBuffer, SizeOf(FDnsLookupBuffer));
  CheckSocketResult(Ord(FDnsLookupHandle = 0), 'WSAAsyncGetHostByAddr');
end;

procedure THostByAddr.InternalStopProbe;
begin
  CheckSocketResult(WSACancelASyncRequest(FDnsLookupHandle), 'WSACancelASyncRequest');
  inherited;
end;

procedure THostByAddr.InternalWndProc(var Msg: TMessage);
var
    Phe     : Phostent;
    DNSName: string;
begin
  if (Msg.Msg <> WM_ASYNCGETHOSTBYNAME) or (msg.wParam <> LongInt(FDnsLookupHandle)) then
    inherited;
  FDnsLookupHandle := 0;
  if Msg.LParamHi<>0 then
    raise ESocketError.CreateFmt(sWindowsSocketError,
      [GetErrorDescription(Msg.LParamHi), Msg.LParamHi, 'WSAAsyncGetHostByAddr']);
  Phe := PHostent(@FDnsLookupBuffer);
  if (phe <> nil) and (Phe^.h_name<>nil) then
  begin
    SetLength(DnsName, StrLen(Phe^.h_name));
    StrCopy(@DnsName[1], Phe^.h_name);
    Params.ParamByName('HostName').AsString:=DNSName;
    DoOnProbe(True);
  end else DoOnError(0, sNoDNSEntryFound);
end;

{ TTCPPortProbe }

procedure TTCPPortProbe.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftInteger, 'Port', ptInput, [poRequired]);
  Params.ParamByName('Timeout').AsInteger:=10000;
  Params.CreateParam(ftBoolean, 'Port open', ptOutput, []);
  Params.CreateParam(ftString, 'PortData', ptOutput, []);
end;

procedure TTCPPortProbe.DoConnect;
const
  WriteStr = #13#10;
begin
  Params.ParamByName('Port open').AsBoolean:=True;
  DoOnProbe(False);
  SocketWrite(FSocket, WriteStr, Sizeof(WriteStr)){: integer};
  SetAsyncStyles(FSocket, [asRead]);
end;

procedure TTCPPortProbe.DoReceive;
begin
  Params.ParamByName('PortData').AsString:=#13#10+SocketReadStr(FSocket, nil)+#13#10;
  DoOnProbe(True);
end;

class function TTCPPortProbe.GetDescription: String;
begin
  result:=sTCPPortProbeDesc;
end;

class function TTCPPortProbe.GetName: string;
begin
  result:='TCP port probe';
end;

procedure TTCPPortProbe.InternalStartSocket;
var
  addr: TSockAddrIn;
begin
  addr.sin_family:=AF_INET;
  addr.sin_port:=htons(Params.ParamByName('Port').AsInteger);
  addr.sin_addr.S_addr:=Params.ParamByName('Address').AsInteger;
  SocketConnect(FSocket, addr);
end;

function TTCPPortProbe.OpenSocket: TSocket;
begin
  result:=GetSocket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
  SetAsyncStyles(Result, [asConnect]);
end;

{ THostAddress }

procedure THostAddress.AddSupportedParams(var Params: TParams);
begin
  inherited;
  Params.CreateParam(ftString, 'Host', ptInput, [poRequired]);
  Params.CreateParam(ftIPAddress, 'Address', ptOutput, []);
end;

destructor THostAddress.Destroy;
begin
  if FHostByName<>nil then FHostByName.Free;
  inherited;
end;

class function THostAddress.GetDescription: String;
begin
  Result:=sHostAddressDesc;
end;

class function THostAddress.GetName: string;
begin
  Result:='Get host address';
end;

procedure THostAddress.HostResolved(Probe: TCustomProbe);
begin
  if Probe.Params.ParamByName(pnProbeSuccess).AsBoolean then
  begin
    Params.ParamByName('Address').AsInteger:=Probe.Params.ParamByName('Address').AsInteger;
    DoOnProbe(True);
  end else
  begin
    DoOnError(Probe.Params.ParamByName('ErrorCode').AsInteger,
      Probe.Params.ParamByName('ErrorMsg').AsString);
  end;
  FHostByName.Free;
  FHostByName:=nil;
end;

function THostAddress.InternalAction: Boolean;

  function IsAddress(const Host: string): Boolean;
  var
    I, Points: Integer;
  begin
    Points := 0;
    for I := 1 to Length(Host) do
    begin
      if (Host[I] = '.') then
      begin
        Inc(Points);
        if Points > 3 then
        begin
          Result := False;
          Exit;
        end else
          Continue;
      end;

      if not (Host[I] in ['0'..'9']) then
      begin
        Result := False;
        Exit;
      end;
    end;
    Result := True;
  end;

begin
  result:=False;
  if IsAddress(Params.ParamByName('Host').AsString) then
  begin
    Params.ParamByName('Address').AsString:=Params.ParamByName('Host').AsString;
    DoOnProbe(True);
  end else
  begin
    FHostByName:=THostByName.Create;
    FHostByName.Params.ParamByName('HostName').AsString:=
      Params.ParamByName('Host').AsString;
    FHostByName.NotifyByMessage:=False;
    FHostByName.Notify:=HostResolved;
    FHostByName.OnIdle:=OnIdle;
    FHostByName.StartProbe;
  end;
end;

initialization
  RegisterProbes([TIPRangeIterator, TPortRangeIterator, THostByName,
    THostByAddr, TPingProbe, TTCPPortProbe, THostAddress]);
finalization
  FreeWinsock;
end.
