{*******************************************************}
{          Copyright (c) 2000 Vlad Korshunov            }
{                   IPH tools project                   }
{              Main probe management unit               }
{*******************************************************}
unit session;
interface
uses
  windows, sysutils, messages, classes, probes, probeparams, dialogs,
  controls, parser;

type
  TTemplateItem = class;

  {store information about parameter sources}
  TParamSource = class
  private
    FParamName: string;
    FSourceParam: TParam;
    FDefaultValue: String;
  public
    constructor Create(const ParamName: String; SourceParam: TParam;
      const DefaultValue: String);
    property ParamName: String read FParamName;
    property DefaultValue: String read FDefaultValue;
    property SourceParam: TParam read FSourceParam;
  end;

  TParamList = class(TList)
  private
    function Get(Index: Integer): TParamSource;
    procedure Put(Index: Integer; Item: TParamSource);
  public
    destructor Destroy; override;
    procedure FreeItems;
    property Items[Index: Integer]: TParamSource read Get write Put; default;
    function FindByParamName(const ParamName: string): TParamSource;
    function FindBySourceParam(Param: TParam): TParamSource;
    function NewItem(const ParamName: String; SourceParam: TParam;
      const DefaultValue: String): TParamSource;
  end;

  TTemplateItemEvent = procedure(Sender: TTemplateItem) of object;
  
  {Template item}
  TTemplateItem = class
  private
    FOwnerItem: TTemplateItem;
    FProbeName: String;                 // Probe name
    FAlias: String;                     // Probe alias
    FChildItems: TList;                 // Child probes
    FParams: TParams;                   // Params supported by probe, not stored
    FParamSources: TParamList;          // Param sources, stored
    FParamsReady: Boolean;
    FOnItemRemoved: TTemplateItemEvent;
    function FindByAlias(const Alias: String): TTemplateItem;
    function GetNewAlias(const Alias: String): String;
    function GetRootItem: TTemplateItem;
    function GetParams: TParams;
    procedure QueryParams;
  public
    constructor Create(OwnerItem: TTemplateItem; ProbeName, Alias: string);
    destructor destroy; override;
    function ParamUsedInChilds(Param: TParam): Boolean;
    function CreateProbe: TCustomProbe;
    procedure AddChild(Item: TTemplateItem; First: Boolean = False);
    procedure RemoveChild(Item: TTemplateItem);
    property ChildItems: TList read FChildItems;
    property ProbeName: string read FProbeName;
    property Alias: String read FAlias write FAlias;
    property OwnerItem: TTemplateItem read FOwnerItem;
    property Params: TParams read GetParams;
    property ParamSources: TParamList read FParamSources;
    property OnItemRemoved: TTemplateItemEvent read FOnItemRemoved write FOnItemRemoved;
  end;

  TGroupList = class;

  TParamGroup = class(TList)
  private
    FOwner: TGroupList;
    FGroupName: String;
    function Get(Index: Integer): TParam;
    procedure Put(Index: Integer; Item: TParam);
    procedure SetGroupName(const Value: String);
  public
    constructor Create(const GroupName: String; Owner: TGroupList);
    function NewItem(Param: TParam): Boolean; // return False if Param Already exists
    property Items[Index: Integer]: TParam read Get write Put; default;
    property GroupName: String read FGroupName write SetGroupName;
  end;

  TGroupList = class(TList)
  private
    function Get(Index: Integer): TParamGroup;
    procedure Put(Index: Integer; Item: TParamGroup);
  public
    destructor Destroy; override;
    function FindByGroupName(const GroupName: string): TParamGroup;
    function NewItem(const GroupName: string): TParamGroup;
    function GetUniqueName(const GroupName: string): String;
    procedure FreeItems;
    property Items[Index: Integer]: TParamGroup read Get write Put; default;
  end;

  TSessionView = class
  private
    FGroups: TGroupList; // 1'st group is Visible params
  public
    constructor Create;
    destructor Destroy; override;
    property Groups: TGroupList read FGroups;
  end;

  TSessionTemplate = class
  private
    FRootItem: TTemplateItem;
    FView: TSessionView;
    FComment: String;
    FOnItemRemoved: TTemplateItemEvent;
    procedure ItemRemoved(Item: TTemplateItem);
  public
    constructor create;
    destructor destroy; override;
    function FullParamName(Param: TParam): String;
    procedure ParseParamName(const FullName: String;
      var Param: TParam; var TemplateItem: TTemplateItem; RaiseException: Boolean = True);
    function GetParamByFullName(const FullName: String;
      RaiseException: Boolean = True): TParam;
    procedure LoadFromStream(S: TStream);
    procedure SaveToStream(S: TStream);
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const  FileName: string);
    property RootItem: TTemplateItem read FRootItem;
    property Comment: String read FComment write FComment;
    property View: TSessionView read FView;
    property OnItemRemoved: TTemplateItemEvent read FOnItemRemoved write FOnItemRemoved;
  end;

  TSession = class;

  TSessionProbe = class
  private
    FProbe: TCustomProbe;
    FTemplateItem: TTemplateItem;
    FOwner: TSession;
    FFinished: Boolean;
    FSuccess: boolean;
    FChildParams: TParamList;
    procedure SetParams(ParentParams: TParamList);
    function GetActive: boolean;
    procedure SetActive(const Value: boolean);
  public
    constructor Create(Owner: TSession; TemplateItem: TTemplateItem;
      ParentParams: TParamList);
    destructor destroy; override;
    property Probe: TCustomProbe read FProbe;
    property TemplateItem: TTemplateItem read FTemplateItem;
    property Active: boolean read GetActive write SetActive;
    property Finished: boolean read FFinished write FFinished;
    property Success: boolean read FSuccess write FSuccess;
    property ChildParams: TParamList read FChildParams;
  end;

  TSetProbesParams = procedure(Session: TSession; var Continue: Boolean) of object;

  TProbeStateChange = procedure(Session: TSession;
    SessionProbe: TSessionProbe) of object;

  TExceptionEvent = procedure(Session: TSession; E: Exception) of object;

  TSession = class(TThread)
  private
    FProbes: TList;
    FTemplate: TSessionTemplate;
    FOnSetProbesParams: TSetProbesParams;
    FOnProbeStateChange: TProbeStateChange;
    FOnException: TExceptionEvent;
    FActiveProbeCount: Integer;
    FMaxProbes: Integer;

    FCurrentSessionProbe: TSessionProbe;
    FCurrentContinue: boolean;
    FCurrentException: Exception;

    procedure CallProbeStateChange;
    procedure CallSetProbeParams;
    procedure CallOnException;

    function FindSessionProbeByProbe(Probe: TCustomProbe): TSessionProbe;

    procedure CreateProbes(TemplateItem: TTemplateItem; Params: TParams;
      ParentParams: TParamList);
    procedure DestroySessionProbes;
    function DoSetProbesParams: Boolean;
    procedure ProbeStateChanged(SessionProbe: TSessionProbe);
    function GetActive: Boolean;
    procedure AddSessionProbe(SessionProbe: TSessionProbe);
    procedure RemoveSessionProbe(SessionProbe: TSessionProbe);
    procedure ProcessMessages;
    function ProcessMessage: Boolean;
    procedure OnProbe(Probe: TCustomProbe; Params: TParams);
    procedure ProbeIdle(Probe: TCustomProbe);
  public
    constructor create(Template: TSessionTemplate);
    procedure Execute; override;
    destructor destroy; override;
    procedure CopyParamList(TemplateItem: TTemplateItem; Source: TParamList; Dest: TParamList);
    property Template: TSessionTemplate read FTemplate;
    property ActiveProbeCount: Integer read FActiveProbeCount;
    property OnSetProbesParams: TSetProbesParams read FOnSetProbesParams write FOnSetProbesParams;
    property OnProbeStateChange: TProbeStateChange read FOnProbeStateChange write FOnProbeStateChange;
    property OnException: TExceptionEvent read FOnException write FOnException;
    property MaxProbes: Integer read FMaxProbes write FMaxProbes;
  end;

  TSessionResult = class
  private
    FTemplateItem: TTemplateItem;
    FParams: TParams;
    FOwnerData: Pointer;
    function GetSuccess: Boolean;
    function GetTime: TDateTime;
    procedure SetSuccess(const Value: Boolean);
    procedure SetTime(const Value: TDateTime);
  public
    constructor Create(TemplateItem: TTemplateItem; Params: TParams; ASuccess: boolean);
    destructor Destroy; override;
    property TemplateItem: TTemplateItem read FTemplateItem;
    property Params: TParams read FParams;
    property Success: Boolean read GetSuccess write SetSuccess;
    property OwnerData: Pointer read FOwnerData write FOwnerData;
    property Time: TDateTime read GetTime write SetTime;
  end;

  TSessionResults = class(TList)
  private
    function Get(Index: Integer): TSessionResult;
    procedure Put(Index: Integer; Item: TSessionResult);
  public
    destructor Destroy; override;
    procedure FreeItems;
    procedure SaveToStream(S: TStream);
// LoadFromStream - only through parser
    property Items[Index: Integer]: TSessionResult read Get write Put; default;
    function NewItem(TemplateItem: TTemplateItem; Params: TParams;
      Success: Boolean): TSessionResult;
  end;

  TTemplateParser = class(TStreamParser)
  private
    FTemplate: TSessionTemplate;
  protected
    function ExpectedStates: TParserStates; override;
    procedure DoState; override;
  public
    constructor Create(Stream: TStream; Template: TSessionTemplate);
  end;

  TSessionParser = class(TTemplateParser)
  private
    FResults: TSessionResults;
  protected
    function ExpectedStates: TParserStates; override;
    procedure DoState; override;
  public
    constructor Create(Stream: TStream; Template: TSessionTemplate;
      Results: TSessionResults);
  end;

function GetParamOwner(Param: TParam): TTemplateItem;

const
  tkResultTimeStamp = '_TimeStamp';
  tkResultSuccess = '_Success';

implementation
uses iphconsts, utils;


function strToken(var S: String; Seperator: Char): String;
var
  I               : Word;
begin
  I:=Pos(Seperator,S);
  if I<>0 then
  begin
    Result:=System.Copy(S,1,I-1);
    System.Delete(S,1,I);
  end else
  begin
    Result:=S;
    S:='';
  end;
end;

{ TTemplateItem }
procedure TTemplateItem.AddChild(Item: TTemplateItem; First: Boolean = False);
begin
  if Item.Alias = '' then Item.Alias:=GetNewAlias(Item.ProbeName);
  if First then FChildItems.Insert(0, Item) else FChildItems.Add(Item);
  Item.FOwnerItem:=Self;
end;

constructor TTemplateItem.Create(OwnerItem: TTemplateItem; ProbeName, Alias: string);
begin
  inherited Create;
  FChildItems:=TList.Create;
  FParams:=TParams.Create(Self);
  FParamSources:=TParamList.Create;
  FOwnerItem:=OwnerItem;
  if FOwnerItem<>nil then FOnItemRemoved:=FOwnerItem.FOnItemRemoved;
  FProbeName:=ProbeName;
  FAlias:=Alias;
end;

procedure TTemplateItem.QueryParams;
var
  i: integer;
  Param: TParam;
  Probe: TCustomProbe;
begin
  if (FParamsReady) or (FOwnerItem = nil) then Exit;
  Probe:=CreateProbe;
  if Probe<>nil then
    With Probe do
    try
      FParams.Assign(Params);
      // Create default source parameters
      for i:=0 to FParams.Count-1 do
      if (FParams[i].ParamType = ptInput) and (FParams[i].IsNull) then
      begin
        Param:=OwnerItem.Params.FindParam(FParams[i].Name);
        if (Param<>nil) and (FParamSources.FindByParamName(FParams[i].Name) = nil) then
          FParamSources.NewItem(FParams[i].Name, Param, '');
      end;
    finally
      Free;
    end;
  FParamsReady:=True;
end;

function TTemplateItem.CreateProbe: TCustomProbe;
begin
  result:=GetProbeByName(ProbeName);
end;

destructor TTemplateItem.destroy;
var
  i: integer;
begin
  for i:=0 to FChildItems.Count-1 do
    TTemplateItem(FChildItems[i]).Free;
  if Assigned(FOnItemRemoved) then FOnItemRemoved(Self);
  FChildItems.Free;
  FParams.Free;
  FParamSources.Free;
  inherited;
end;

procedure TTemplateItem.RemoveChild(Item: TTemplateItem);
begin
  FChildItems.Remove(Item);
  Item.FOwnerItem:=nil;
end;

function TTemplateItem.FindByAlias(const Alias: String): TTemplateItem;
var
  i: Integer;
begin
  result:=nil;
  if AnsiUpperCase(FAlias) <> AnsiUpperCase(Alias) then
    for i:=0 to ChildItems.Count-1 do
    begin
      result:=TTemplateItem(ChildItems[i]).FindByAlias(Alias);
      if result<>nil then Break;
    end else
      result:=Self;
end;

function TTemplateItem.GetNewAlias(const Alias: String): String;
var
  FRootItem: TTemplateItem;
  i: integer;
begin
  FRootItem:=GetRootItem;
  i:=1;
  result:=Alias;
  while FRootItem.FindByAlias(result)<>nil do
  begin
    result:=Alias+IntToStr(i);
    inc(i);
  end;
end;

function TTemplateItem.GetRootItem: TTemplateItem;
begin
  Result:=Self;
  While Result.FOwnerItem<>nil do
    Result:=Result.FOwnerItem;
end;

function TTemplateItem.ParamUsedInChilds(Param: TParam): Boolean;

  function OneItem(Item: TTemplateItem): Boolean;
  var
    i: integer;
  begin
    result:=False;
    for i:=0 to Item.ParamSources.Count-1 do
    begin
      result:=False;
      if (Item.ParamSources[i].SourceParam = Param) then
      begin
        result:=True;
        Break;
      end;
    end;
    if not result then
      for i:=0 to Item.ChildItems.Count-1 do
      begin
        result:=OneItem(Item.ChildItems[i]);
        if result then break;
      end;
  end;

begin
  result:=OneItem(Self);
end;

function TTemplateItem.GetParams: TParams;
begin
  QueryParams;
  result:=FParams;
end;

{ TSessionTemplate }

constructor TSessionTemplate.create;
begin
  inherited Create;
  FRootItem:=TTemplateItem.Create(nil, 'ROOT', sRootAlias);
  FRootItem.OnItemRemoved:=ItemRemoved;
  FView:=TSessionView.Create;
end;

destructor TSessionTemplate.destroy;
begin
  FRootItem.Free;
  FView.Free;;
  inherited;
end;

function TSessionTemplate.FullParamName(Param: TParam): String;
begin
  result:=GetParamOwner(Param).Alias+'.'+Param.Name;
end;

function TSessionTemplate.GetParamByFullName(
  const FullName: String; RaiseException: Boolean = True): TParam;
var
  TemplateItem: TTemplateItem;
begin
  ParseParamName(FullName, Result, TemplateItem, RaiseException);
end;

procedure TSessionTemplate.ItemRemoved(Item: TTemplateItem);
var
  i, j, k: integer;
begin
  // Remove item params from view groups
  for k:=0 to Item.Params.Count-1 do
    for i:=0 to FView.Groups.Count-1 do
      for j:=FView.Groups[i].Count-1 downto 0 do
        if FView.Groups[i][j] = Item.Params[k] then
           FView.Groups[i].Remove(FView.Groups[i][j]);
  if Assigned(FOnItemRemoved) then FOnItemRemoved(Item);
end;

procedure TSessionTemplate.LoadFromFile(const FileName: string);
var
  FileStream: TFileStream;
begin
  FileStream:=TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

procedure TSessionTemplate.LoadFromStream(S: TStream);
begin
  with TTemplateParser.Create(S, Self) do
  try
    Parse;
  finally
    Free;
  end;
end;

procedure TSessionTemplate.ParseParamName(const FullName: String;
  var Param: TParam; var TemplateItem: TTemplateItem;
  RaiseException: Boolean = True);

  function FindTemplateItem(Alias: String): TTemplateItem;
  begin
    result:=nil;
    if Alias<>'' then
    begin
      result:=FRootItem.FindByAlias(Alias);
      if (result = nil) and RaiseException then
        Raise exception.CreateFmt(sInvalidAlias, [Alias]);
    end;
  end;

var
  s: string;

begin
  s:=FullName;
  TemplateItem:=FindTemplateItem(StrToken(S, '.'));
  if TemplateItem<>nil then
  begin
    Param:=TemplateItem.Params.FindParam(S);
    if (Param = nil) and RaiseException then
      Raise exception.CreateFmt(sInvalidParameter, [FullName]);
  end else Param:=nil;
end;

procedure TSessionTemplate.SaveToFile(const FileName: string);
var
  FileStream: TFileStream;
begin
  FileStream:=TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(FileStream);
  finally
    FileStream.Free;
  end;
end;

procedure TSessionTemplate.SaveToStream(S: TStream);
var
  Level: integer;

  procedure WriteParams(Item: TTemplateItem);
  var
    i: integer;
    str: string;
  begin
    if Item.Paramsources.Count>0 then WriteStrLevel(S, Level, StartTokens[stParamSource]);
    for i:=0 to Item.Paramsources.Count-1 do
    With TParamSource(Item.Paramsources[i]) do
    begin
      if SourceParam<>nil then
        str:=ParamName+'='+FullParamName(SourceParam) else
        str:=ParamName+'='+DefaultValue;
      // skip default params
      if (SourceParam <> nil) or (DefaultValue <> Item.Params.ParamByName(ParamName).AsString) then
        WriteStrLevel(S, Level+1, str);
    end;
    if Item.Paramsources.Count>0 then WriteStrLevel(S,Level, StopTokens[stParamSource]);
  end;

  procedure RecurseWrite(Items: TList);
  var
    i: Integer;
  begin
    for i:=0 to Items.Count-1 do
    begin
      With TTemplateItem(Items[i]) do
        WriteStrLevel(S, Level, ProbeName+Seperator+Alias);
      Inc(Level);
      WriteParams(Items[i]);
      RecurseWrite(TTemplateItem(Items[i]).ChildItems);
      Dec(Level);
    end;
  end;

  procedure WriteSessionView;
  var
    i, j: integer;
  begin
    for i:=0 to FView.FGroups.Count-1 do
    begin
      WriteStrLevel(S, 1, FView.FGroups[i].FGroupName);
      WriteStrLevel(S, 2, StartTokens[stGroupParam]);
      for j:=0 to FView.FGroups[i].Count-1 do
        WriteStrLevel(S, 3, FullParamName(FView.FGroups[i][j]));
      WriteStrLevel(S, 2, StopTokens[stGroupParam]);
    end;
  end;

begin
  WriteStateBegin(S, stTreeItem);
  Level:=1;
  RecurseWrite(FRootItem.ChildItems);
  WriteStateEnd(S, stTreeItem);
  WriteStateBegin(S, stViewGroup);
  WriteSessionView;
  WriteStateEnd(S, stViewGroup);
  WriteStateBegin(S, stComment);
  StreamWriteStr(S, FComment+EndOfLine);
  WriteStateEnd(S, stComment);
end;

{ TSession }

constructor TSession.create(Template: TSessionTemplate);
begin
  inherited Create(True);
  FMaxProbes:=30;
  FProbes:=TList.Create;
  FTemplate:=Template;
end;

destructor TSession.destroy;
begin
  FProbes.Free;
  inherited;
end;

procedure TSession.Execute;
begin
  try
    try
      // set root probes params
      if DoSetProbesParams then
        CreateProbes(FTemplate.FRootItem, nil, nil);
      repeat
        ProcessMessages;
      until terminated;
    finally
      DestroySessionProbes;
    end;
  except
    On E: Exception do
    begin
      FCurrentException:=E;
      Synchronize(CallOnException);
    end;
  end;
end;

procedure TSession.CallProbeStateChange;
begin
  if Assigned(FOnProbeStateChange) then
    FOnProbeStateChange(Self, FCurrentSessionProbe);
end;

procedure TSession.CallSetProbeParams;
begin
  if Assigned(FOnSetProbesParams) then
    FOnSetProbesParams(Self, FCurrentContinue);
end;

procedure TSession.CallOnException;
begin
  if Assigned(FOnException) then FOnException(Self, FCurrentException);
end;

function TSession.DoSetProbesParams: boolean;
begin
  FCurrentContinue:=True;
  Synchronize(CallSetProbeParams);
  result:=FCurrentContinue;
end;

procedure TSession.ProbeStateChanged(SessionProbe: TSessionProbe);
begin
  FCurrentSessionProbe:=SessionProbe;
  Synchronize(CallProbeStateChange);
end;

procedure TSession.DestroySessionProbes;
var
  i: integer;
begin
  for i:=0 to FProbes.Count-1 do
    TSessionProbe(FPRobes[i]).Free;
end;

function TSession.GetActive: Boolean;
var
  i: integer;
begin
  FActiveProbeCount:=0;
  result:=False;
  for i:=0 to FProbes.Count-1 do
    if TSessionProbe(FProbes[i]).Active or not TSessionProbe(FProbes[i]).Finished then
    begin
      inc(FActiveProbeCount);
      result:=True;
    end;
end;

function TSession.ProcessMessage: Boolean;
var
  msg: TMsg;
begin
  result:=
    PeekMessage(Msg, 0, WM_PROBE_EVENT, WM_PROBE_EVENT, PM_REMOVE) or
    PeekMessage(Msg, 0, 0, 0, PM_REMOVE);
  if result then
  Case Msg.Message of
    WM_QUIT:
      begin
        Terminate;
        result:=False;
      end;
    WM_PROBE_EVENT: OnProbe(TCustomProbe(Pointer(Msg.WParam)),
      TParams(Pointer(Msg.LParam)));
    else
      begin
        TranslateMessage(Msg);
        DispatchMessage(Msg);
      end;
  end;
end;

procedure TSession.ProcessMessages;
begin
  while ProcessMessage do ;
end;

procedure TSession.AddSessionProbe(SessionProbe: TSessionProbe);
var
  i: integer;
begin
  while FProbes.Count>=FMaxProbes do
  begin
    ProcessMessage;
    for i:=FProbes.Count-1 downto 0 do
      if (not TSessionProbe(FProbes[i]).Active) and TSessionProbe(FProbes[i]).Finished
        then RemoveSessionProbe(FProbes[i]);
  end;
  FProbes.Add(SessionProbe);
end;

procedure TSession.RemoveSessionProbe(SessionProbe: TSessionProbe);
begin
  FProbes.Remove(SessionProbe);
  SessionProbe.Free;
end;

function TSession.FindSessionProbeByProbe(
  Probe: TCustomProbe): TSessionProbe;
var
  i: integer;
begin
  result:=nil;
  for i:=0 to FProbes.Count-1 do
    if TSessionProbe(FProbes[i]).Probe = Probe then
    begin
      result:=TSessionProbe(FProbes[i]);
      break;
    end;
end;

procedure TSession.OnProbe(Probe: TCustomProbe; Params: TParams);
var
  SessionProbe: TSessionProbe;
  TemplateItem: TTemplateItem;
  ParentParams: TParamList;
  FSuccess: Boolean;
  FFinished: Boolean;
begin
  SessionProbe:=FindSessionProbeByProbe(Probe);
  Assert(SessionProbe<>nil);
  FSuccess:=Params.ParamByName('ProbeSuccess').AsBoolean;
  FFinished:=Params.ParamByName('ProbeFinished').AsBoolean;
  SessionProbe.Success:=FSuccess;
  if FFinished then ProbeStateChanged(SessionProbe);
  TemplateItem:=SessionProbe.TemplateItem;
  ParentParams:=TParamList.Create;
  try
    CopyParamList(TemplateItem, SessionProbe.FChildParams, ParentParams);
    SessionProbe.Finished:=FFinished; // After this Session Probe may be destroyed
    if FSuccess then
      CreateProbes(TemplateItem, Params, ParentParams);
    if not GetActive then terminate;
  finally
    ParentParams.Free;
  end;
end;

procedure TSession.CreateProbes(TemplateItem: TTemplateItem; Params: TParams;
  ParentParams: TParamList);

  procedure CopyParams;
  var
    i: Integer;
    TemplateParam: TParam;
  begin
    // Transfer values used in childs from params to ParentParams
    for i:=0 to Params.Count-1 do
    begin
      TemplateParam:=TemplateItem.Params.FindParam(Params[i].Name);
      if TemplateItem.ParamUsedInChilds(TemplateParam) then
        ParentParams.NewItem(Params[i].Name, TemplateParam, Params[i].AsString);
    end;
  end;

var
  NewSessionProbe: TSessionProbe;
  i: integer;
  created: boolean;
begin
  created:=ParentParams=nil;
  if created then ParentParams:=TParamList.Create;
  try
    if Params<>nil then CopyParams;
    for i:=0 to TemplateItem.ChildItems.Count-1 do
    begin
      if Terminated then break;
        NewSessionProbe:=TSessionProbe.Create(Self, TTemplateItem(TemplateItem.ChildItems[i]),
          ParentParams);
        NewSessionProbe.Probe.OnIdle:=ProbeIdle;
      AddSessionProbe(NewSessionProbe);
      NewSessionProbe.Active:=True;
      if not NewSessionProbe.Active then break;
    end;
  finally
    if created then ParentParams.Free;
  end;
end;

procedure TSession.CopyParamList(TemplateItem: TTemplateItem; Source,
  Dest: TParamList);
var
  i: integer;
begin
  for i:=0 to Source.Count-1 do
  with Source[i] do
  begin
    if TemplateItem.ParamUsedInChilds(SourceParam) then
      Dest.NewItem(ParamName, SourceParam, DefaultValue);
  end;
end;

procedure TSession.ProbeIdle(Probe: TCustomProbe);
begin
  ProcessMessage;
end;

{ TSessionProbe }

constructor TSessionProbe.Create(Owner: TSession;
  TemplateItem: TTemplateItem; ParentParams: TParamList);
begin
  inherited Create;
  FOwner:=Owner;
  FChildParams:=TParamList.Create;
  FTemplateItem:=TemplateItem;
  Assert(TemplateItem<>nil);
  FProbe:=TemplateItem.CreateProbe;
  SetParams(ParentParams); // must be last
end;

destructor TSessionProbe.destroy;
begin
  Active:=False;
  FProbe.Free;
//  FChildParams.Free; - destroyed by owner
  inherited;
end;

function TSessionProbe.GetActive: boolean;
begin
  result:=FProbe.Active;
end;

procedure TSessionProbe.SetActive(const Value: boolean);
begin
  if Active<>Value then
  begin
    if Value then Probe.StartProbe else Probe.StopProbe;
    FOwner.ProbeStateChanged(Self);
  end;
end;

procedure TSessionProbe.SetParams(ParentParams: TParamList);

  procedure TransferChildParams;
  // transfer input or output params from parent to child item(self), because
  // parent destroyed, and childs cannot access params
  var
    i: integer;
  begin
    for i:=0 to ParentParams.Count-1 do
    with ParentParams[i] do
    begin
      if FTemplateItem.ParamUsedInChilds(SourceParam) then
        FChildParams.NewItem(ParamName, SourceParam, DefaultValue);
    end;
  end;

var
  i: integer;
  ParamSource: TParamSource;
  ParamName: string;
begin
  if ParentParams<>nil then
  begin
    TransferChildParams;
    for i:=0 to FProbe.Params.Count-1 do
      if (FProbe.Params[i].ParamType=ptInput) then
      begin
        ParamName:=FProbe.Params[i].Name;
        ParamSource:=FTemplateItem.ParamSources.FindByParamName(FProbe.Params[i].Name);

        // Sourced param
        if (ParamSource<>nil) and (ParamSource.SourceParam<>nil) then
          ParamSource:=ParentParams.FindBySourceParam(ParamSource.SourceParam);

        if (ParamSource<>nil) and (ParamSource.DefaultValue<>'') then
          FProbe.Params[i].AsString:=ParamSource.FDefaultValue;
      end;
  end;
end;

{ TParamList }

destructor TParamList.Destroy;
begin
  FreeItems;
  inherited;
end;

function TParamList.FindByParamName(const ParamName: string): TParamSource;
var
  i: integer;
begin
  result:=nil;
  for i:=0 to Count-1 do
    if (AnsiUpperCase(Items[i].FParamName) = AnsiUpperCase(ParamName)) then
    begin
      Result:=Items[i];
      break;
    end;
end;

function TParamList.FindBySourceParam(Param: TParam): TParamSource;
var
  i: integer;
begin
  result:=nil;
  for i:=0 to Count-1 do
    if (Items[i].SourceParam = Param) then
    begin
      Result:=Items[i];
      break;
    end;
end;

procedure TParamList.FreeItems;
var
  i: Integer;
begin
  for i:=0 to Count-1 do Items[i].Free;
end;

function TParamList.Get(Index: Integer): TParamSource;
begin
  result:=TParamSource(inherited Items[Index]);
end;

function TParamList.NewItem(const ParamName: String; SourceParam: TParam;
  const DefaultValue: String): TParamSource;
begin
  result:=FindByParamName(ParamName);
  if result = nil then
    result:=Items[Add(TParamSource.Create(ParamName, SourceParam, DefaultValue))] else
  begin
    result.FParamName:=ParamName;
    result.FSourceParam:=SourceParam;
    result.FDefaultValue:=DefaultValue;
  end;
end;

procedure TParamList.Put(Index: Integer; Item: TParamSource);
begin
  inherited Items[Index]:=Item;
end;

{ TParamSource }

constructor TParamSource.Create(const ParamName: String; SourceParam: TParam;
  const DefaultValue: String);
begin
  inherited Create;
  FParamName:=ParamName;
  FSourceParam:=SourceParam;
  FDefaultValue:=DefaultValue;
end;

{ TSessionResults }

destructor TSessionResults.Destroy;
begin
  FreeItems;
  inherited;
end;

procedure TSessionResults.FreeItems;
var
  i: integer;
begin
  for i:=0 to Count-1 do Items[i].Free;
end;

function TSessionResults.Get(Index: Integer): TSessionResult;
begin
  result:=inherited Items[Index];
end;

function TSessionResults.NewItem(TemplateItem: TTemplateItem;
  Params: TParams; Success: Boolean): TSessionResult;
begin
  result:=Items[Add(TSessionResult.Create(TemplateItem, Params, Success))];
end;

procedure TSessionResults.Put(Index: Integer; Item: TSessionResult);
begin
  inherited Items[Index]:=Item;
end;

procedure TSessionResults.SaveToStream(S: TStream);

  function IsComplexParam(Param: TParam): Boolean;
  begin
    result:=Pos(#13, Param.AsString)>0;
  end;

  procedure WriteValue(const Value: String);
  begin
    WriteStrLevel(S, 4, StartTokens[stParamValue]);
    StreamWriteStr(S, Value);
    WriteStrLevel(S, 4, StopTokens[stParamValue]);
  end;

  procedure WriteParam(Param: TParam);
  var
    str: String;
  begin
    if Param.AsString<>'' then
    begin
      str:=Format('%s;%d;%d=',[Param.Name, Integer(Param.DataType),
        Integer(Param.ParamType)]);
      if IsComplexParam(Param) then
      begin
        WriteStrLevel(S, 3, str);
        WriteValue(Param.AsString);
      end else
      begin
        str:=Str+Param.AsString;
        WriteStrLevel(S, 3, str);
      end;
    end;
  end;

var
  i, j: integer;
begin
  WriteStateBegin(S, stResult);
  for i:=0 to Count-1 do
  begin
    WriteStrLevel(S, 1, Items[i].TemplateItem.Alias);
    if Items[i].Params.Count<>0 then
    begin
      WriteStrLevel(S, 2, StartTokens[stResultParams]);
      for j:=0 to Items[i].Params.Count-1 do
        WriteParam(Items[i].Params[j]);
      WriteStrLevel(S, 2, StopTokens[stResultParams])
    end;
  end;
  WriteStateEnd(S, stResult);
end;

{ TSessionResult }

constructor TSessionResult.Create(TemplateItem: TTemplateItem;
  Params: TParams; ASuccess: boolean);
begin
  inherited Create;
  FTemplateItem:=TemplateItem;
  FParams:=TParams.Create;
  if Params<>nil then FParams.Assign(Params);
  Success:=ASuccess;
  Time:=Now;
end;

destructor TSessionResult.Destroy;
begin
  FParams.Free;
  inherited;
end;

function TSessionResult.GetSuccess: Boolean;
begin
  result:=Params.ParamByName(tkResultSuccess).AsBoolean;
end;

function TSessionResult.GetTime: TDateTime;
begin
  result:=Params.ParamByName(tkResultTimeStamp).AsDateTime;
end;

procedure TSessionResult.SetSuccess(const Value: Boolean);
var
  Param: TParam;
begin
  Param:=Params.FindParam(tkResultSuccess);
  if Param=nil then Param:=Params.CreateParam(ftBoolean, tkResultSuccess,
    ptOutput, []);
  Param.AsBoolean:=Value;
end;

procedure TSessionResult.SetTime(const Value: TDateTime);
var
  Param: TParam;
begin
  Param:=Params.FindParam(tkResultTimeStamp);
  if Param=nil then Param:=Params.CreateParam(ftDateTime, tkResultTimeStamp,
    ptOutput, []);
  Param.AsDateTime:=Value;
end;

{ TSessionView }

constructor TSessionView.Create;
begin
  inherited Create;
  FGroups:=TGroupList.Create;
  // Visible params group always exists
  FGroups.NewItem(grVisibleParams);
end;

destructor TSessionView.Destroy;
begin
  FGroups.Free;
  inherited;
end;

{ TParamGroup }

constructor TParamGroup.Create(const GroupName: String; Owner: TGroupList);
begin
  inherited Create;
  FOwner:=Owner;
  FGroupName:=GroupName;
end;

function TParamGroup.Get(Index: Integer): TParam;
begin
  result:=inherited Items[Index];
end;

function TParamGroup.NewItem(Param: TParam): Boolean;
var
  Index: Integer;
begin
  result:=True;
  index:=IndexOf(Param);
  if Index<>-1 then result:=False else Add(Param);
end;

procedure TParamGroup.Put(Index: Integer; Item: TParam);
begin
  inherited Items[Index]:=Item;
end;

procedure TParamGroup.SetGroupName(const Value: String);
begin
  if FOwner.FindByGroupName(Value)<>nil then
    Raise Exception.Create(sGroupNameAlreadyExists);
  FGroupName := Value;
end;

{ TGroupList }

destructor TGroupList.Destroy;
begin
  FreeItems;
  inherited;
end;

function TGroupList.FindByGroupName(const GroupName: string): TParamGroup;
var
  i: integer;
begin
  result:=nil;
  for i:=0 to Count-1 do
    if Items[i].FGroupName = GroupName then
    begin
      result:=Items[i];
      Break;
    end;
end;

procedure TGroupList.FreeItems;
var
  i: integer;
begin
  for i:=0 to Count-1 do Items[i].Free;
end;

function TGroupList.Get(Index: Integer): TParamGroup;
begin
  result:=inherited Items[Index];
end;

function TGroupList.GetUniqueName(const GroupName: string): String;
var
  i: integer;
begin
  i:=1;
  result:=GroupName;
  while FindByGroupName(result)<>nil do
  begin
    result:=GroupName+IntToStr(i);
    inc(i);
  end;
end;

function TGroupList.NewItem(const GroupName: string): TParamGroup;
begin
  result:=FindByGroupName(GroupName);
  if result=nil then
  begin
    result:=TParamGroup.Create(GroupName, Self);
    Add(result);
  end;
end;

procedure TGroupList.Put(Index: Integer; Item: TParamGroup);
begin
  inherited Items[Index]:=Item;
end;

function GetParamOwner(Param: TParam): TTemplateItem;
var
  Params: TParams;
begin
  Assert(Param.Collection<>nil);
  Params:=Param.Collection as TParams;
  Assert(Params.Owner<>nil);
  result:=Params.Owner as TTemplateItem;
end;

{ TTemplateParser }

constructor TTemplateParser.Create(Stream: TStream;
  Template: TSessionTemplate);
begin
  inherited Create(Stream);
  FTemplate:=Template;
end;

procedure TTemplateParser.DoState;

  Function GetItem(ParentItem: TTemplateItem; s: String): TTemplateItem;
  begin
    Result:=TTemplateItem.Create(ParentItem, StrToken(s, Seperator), StrToken(s, Seperator));
  end;

  procedure DoParam(Item: TTemplateItem; s: String);
  var
    Source, ParamName: String;
    SourceParam: TParam;
  begin
    ParamName:=StrToken(s, '=');
    Source:=StrToken(s, '=');
    SourceParam:=FTemplate.GetParamByFullName(Source, False);
    if SourceParam = nil then
      Item.FParamSources.NewItem(ParamName, nil, Source) else
      Item.FParamSources.NewItem(ParamName, SourceParam, '');
  end;

  function FindItemByLevel(Level: Integer): TTemplateItem;
  var
    FLevel: Integer;
  begin
    FLevel:=1;
    Result:=FTemplate.RootItem;
    while FLevel<Level do
    begin
      if result.ChildItems.Count=0 then
        Raise Exception.CreateFmt(sInvalidItemLevel, [Trim(CurLine)]);
      result:=result.ChildItems[result.ChildItems.Count-1];
      inc(FLevel);
    end;
  end;
var
  Item: TTemplateItem;
begin
  try
    case state of
      stTreeItem:
        begin
          Item:=FindItemByLevel(GetLevel);
          Item.ChildItems.Add(GetItem(Item, Trim(CurLine)));
        end;
      stParamSource:
        DoParam(FindItemByLevel(GetLevel-1), Trim(CurLine));
      stViewGroup:
        FTemplate.View.Groups.NewItem(Trim(CurLine));
      stGroupParam:
        with FTemplate.View.Groups do
          Items[Count-1].NewItem(FTemplate.GetParamByFullName(Trim(CurLine)));
      stComment:
        begin
          if FTemplate.Comment<>'' then FTemplate.Comment:=FTemplate.Comment+EndOfLine;
          FTemplate.Comment:=FTemplate.Comment+Trim(CurLine);
        end;
    end;
  except
    On E: Exception do if MessageDlg(E.Message+sTerminateLoading,
      mtWarning, [mbYes, mbNo], 0) = mrYes then abort;
  end;
end;

function TTemplateParser.ExpectedStates: TParserStates;
begin
  result:=[];
  case State of
    stNone: result:=[stTreeItem, stViewGroup, stComment];
    stTreeItem: result:=[stTreeItem, stParamSource, stViewGroup, stComment];
    stParamSource: result:=[stTreeItem, stParamSource, stViewGroup, stComment];
    stViewGroup: result:=[stGroupParam, stComment];
    stGroupParam: result:=[stComment];
    stComment: result:=[stComment, stViewGroup];
  end;
end;

{ TSessionParser }

constructor TSessionParser.Create(Stream: TStream;
  Template: TSessionTemplate; Results: TSessionResults);
begin
  inherited Create(Stream, Template);
  FResults:=Results;
end;

procedure TSessionParser.DoState;

  function GetResult: TSessionResult;
  begin
    if FResults.Count=0 then
      Raise Exception.Create(sResultParamNoParent);
    result:=FResults[FResults.Count-1];
  end;

  procedure DoParam(Result: TSessionResult; s: String);
  var
    ParamName: String;
    FieldType: TFieldType;
    ParamType: TParamType;
  begin
    ParamName:=StrToken(s, seperator);
    FieldType:=TFieldType(StrToInt(StrToken(s, seperator)));
    ParamType:=TParamType(StrToInt(StrToken(s, '=')));
    if ParamName = tkResultTimeStamp then Result.Time:=StrToDateTime(s) else
      if ParamName = tkResultSuccess then
        Result.Params.ParamByName(tkResultSuccess).AsString:=s else
    Result.Params.CreateParam(FieldType, ParamName, ParamType, []).AsString:=s;
  end;

  procedure DoValue(Result: TSessionResult; S: String);
  begin
    if Result.Params.Count=0 then
      Raise Exception.Create(sParamVlaueNotParent);
    Result.Params[Result.Params.Count-1].AsString:=
      Result.Params[Result.Params.Count-1].AsString+s+EndOfLine;
  end;


begin
  inherited;
  case state of
    stResult: FResults.NewItem(FTemplate.RootItem.FindByAlias(Trim(CurLine)), nil, False);
    stResultParams:
      DoParam(GetResult, Trim(CurLine));
    stParamValue:
      DoValue(GetResult, Trim(CurLine));
  end;
end;

function TSessionParser.ExpectedStates: TParserStates;
begin
  result:=inherited ExpectedStates;
  case State of
    stNone: result:=result+[stResult];
    stTreeItem: result:=result+[stResult];
    stParamSource: result:=result+[stResult];
    stViewGroup: result:=result+[stResult];
    stGroupParam: result:=result+[stResult];
    stComment: result:=result+[stResult];
    stResult: result:=[stResultParams];
    stResultParams: result:=[stParamValue];
    stParamValue: ;
  end;
end;

initialization
end.
